package ru.dbotthepony.kstarbound.server

import com.github.benmanes.caffeine.cache.CacheLoader
import com.github.benmanes.caffeine.cache.Caffeine
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.SupervisorJob
import kotlinx.coroutines.async
import kotlinx.coroutines.cancel
import kotlinx.coroutines.future.asCompletableFuture
import kotlinx.coroutines.future.await
import kotlinx.coroutines.launch
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kstarbound.math.vector.Vector3i
import ru.dbotthepony.kstarbound.Globals
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.WorldID
import ru.dbotthepony.kstarbound.defs.world.AsteroidsWorldParameters
import ru.dbotthepony.kstarbound.defs.world.FloatingDungeonWorldParameters
import ru.dbotthepony.kstarbound.defs.world.TerrestrialWorldParameters
import ru.dbotthepony.kstarbound.defs.world.WorldTemplate
import ru.dbotthepony.kstarbound.json.readJsonElement
import ru.dbotthepony.kstarbound.json.readJsonObject
import ru.dbotthepony.kstarbound.json.writeJsonElement
import ru.dbotthepony.kstarbound.json.writeJsonObject
import ru.dbotthepony.kstarbound.network.packets.clientbound.UniverseTimeUpdatePacket
import ru.dbotthepony.kstarbound.server.world.LegacyWorldStorage
import ru.dbotthepony.kstarbound.server.world.ServerUniverse
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.server.world.ServerSystemWorld
import ru.dbotthepony.kstarbound.server.world.WorldStorage
import ru.dbotthepony.kstarbound.util.BlockableEventLoop
import ru.dbotthepony.kstarbound.util.JVMClock
import ru.dbotthepony.kstarbound.util.asStringOrNull
import ru.dbotthepony.kstarbound.util.random.random
import ru.dbotthepony.kstarbound.util.toStarboundString
import ru.dbotthepony.kstarbound.util.uuidFromStarboundString
import ru.dbotthepony.kstarbound.world.UniversePos
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.File
import java.sql.DriverManager
import java.util.UUID
import java.util.concurrent.CompletableFuture
import java.util.concurrent.TimeUnit
import java.util.concurrent.locks.ReentrantLock

sealed class StarboundServer(val root: File) : BlockableEventLoop("Server thread") {
	private fun makedir(file: File) {
		if (!file.exists()) {
			check(file.mkdirs()) { "Unable to create ${file.absolutePath}" }
		} else if (!file.isDirectory) {
			throw IllegalArgumentException("${file.absolutePath} is not a directory")
		}
	}

	val universeFolder = File(root, "universe")

	init {
		makedir(root)
		makedir(universeFolder)
	}

	private val worlds = HashMap<WorldID, CompletableFuture<ServerWorld>>()
	val universe = ServerUniverse(universeFolder)
	val chat = ChatHandler(this)
	val globalScope = CoroutineScope(Starbound.COROUTINE_EXECUTOR + SupervisorJob())

	private val database = DriverManager.getConnection("jdbc:sqlite:${File(universeFolder, "universe.db").absolutePath.replace('\\', '/')}")

	init {
		database.autoCommit = false

		database.createStatement().use {
			it.execute("CREATE TABLE IF NOT EXISTS `metadata` (`key` VARCHAR NOT NULL PRIMARY KEY, `value` BLOB NOT NULL)")
			it.execute("CREATE TABLE IF NOT EXISTS `universe_flags` (`flag` VARCHAR NOT NULL PRIMARY KEY)")
			it.execute("CREATE TABLE IF NOT EXISTS `client_context` (`uuid` VARCHAR NOT NULL PRIMARY KEY, `data` BLOB NOT NULL)")
		}
	}

	private val lookupMetadata = database.prepareStatement("SELECT `value` FROM `metadata` WHERE `key` = ?")
	private val writeMetadata = database.prepareStatement("REPLACE INTO `metadata` (`key`, `value`) VALUES (?, ?)")
	private val lookupClientContext = database.prepareStatement("SELECT `data` FROM `client_context` WHERE `uuid` = ?")
	private val writeClientContext = database.prepareStatement("REPLACE INTO `client_context` (`uuid`, `data`) VALUES (?, ?)")

	private fun getMetadata(key: String): KOptional<JsonElement> {
		lookupMetadata.setString(1, key)

		return lookupMetadata.executeQuery().use {
			if (it.next()) {
				KOptional(it.getBytes(1).readJsonElement())
			} else {
				KOptional()
			}
		}
	}

	private fun setMetadata(key: String, value: JsonElement) {
		writeMetadata.setString(1, key)
		writeMetadata.setBytes(2, value.writeJsonElement())
		writeMetadata.execute()
	}

	fun loadClientContext(uuid: UUID): CompletableFuture<JsonObject?> {
		return supplyAsync {
			lookupClientContext.setString(1, uuid.toStarboundString())

			lookupClientContext.executeQuery().use {
				if (it.next()) {
					it.getBytes(1).readJsonObject()
				} else {
					null
				}
			}
		}
	}

	fun writeClientContext(uuid: UUID, context: JsonObject): CompletableFuture<*> {
		return supplyAsync {
			writeClientContext.setString(1, uuid.toStarboundString())
			writeClientContext.setBytes(2, context.writeJsonObject())
			writeClientContext.execute()
		}
	}

	val settings = ServerSettings()
	val channels = ServerChannels(this)
	val lock = ReentrantLock()
	var isClosed = false
		private set

	val serverUUID: UUID = uuidFromStarboundString(getMetadata("server_uuid").orElse { JsonPrimitive(UUID.randomUUID().toStarboundString()) }.asString)
	val universeClock = JVMClock()

	init {
		universeClock.set(getMetadata("universe_clock").orElse { JsonPrimitive(0.0) }.asDouble)
		setMetadata("server_uuid", JsonPrimitive(serverUUID.toStarboundString()))
		database.commit()
	}

	private val systemWorlds = HashMap<Vector3i, CompletableFuture<ServerSystemWorld?>>()

	private suspend fun loadSystemWorld0(location: Vector3i): ServerSystemWorld? {
		try {
			return ServerSystemWorld.create(this, location)
		} catch (err: Throwable) {
			LOGGER.error("Exception loading system world at $location", err)
			return null
		}
	}

	fun loadSystemWorld(location: Vector3i): CompletableFuture<ServerSystemWorld?> {
		return supplyAsync {
			systemWorlds.computeIfAbsent(location) {
				globalScope.async { loadSystemWorld0(location) }.asCompletableFuture()
			}
		}.thenCompose { it }
	}

	private suspend fun loadCelestialWorld(location: WorldID.Celestial): ServerWorld {
		val fileName = location.pos.toString().replace(':', '_') + ".db"
		val file = File(universeFolder, fileName)
		val firstTime = !file.exists()
		val storage = LegacyWorldStorage.SQL(file)

		val world = try {
			if (firstTime) {
				LOGGER.info("Creating celestial world $location")
				ServerWorld.create(this, WorldTemplate.create(location.pos, universe), storage, location)
			} else {
				LOGGER.info("Loading celestial world $location")
				ServerWorld.load(this, storage, location).await()
			}
		} catch (err: Throwable) {
			storage.close()

			if (firstTime) {
				file.delete()
				throw err
			} else {
				LOGGER.fatal("Exception loading celestial world at $location, recreating!")

				var i = 0
				while (!file.renameTo(File(universeFolder, "$fileName-fail$i")) && ++i < 1000) {}

				ServerWorld.create(this, WorldTemplate.create(location.pos, universe), LegacyWorldStorage.SQL(file), location)
			}
		}

		try {
			world.sky.referenceClock = universeClock
			world.eventLoop.start()
			world.prepare(firstTime).await()

			if (firstTime) {
				world.saveMetadata()
			}
		} catch (err: Throwable) {
			LOGGER.fatal("Exception while initializing celestial world at $location!", err)
			world.eventLoop.shutdown()
			throw err
		}

		return world
	}

	private suspend fun loadInstanceWorld(location: WorldID.Instance): ServerWorld {
		val config = Globals.instanceWorlds[location.name] ?: throw NoSuchElementException("No such instance world ${location.name}")
		LOGGER.info("Creating instance world $location")
		val random = random(config.seed ?: System.nanoTime())

		val visitable = when (config.type.lowercase()) {
			"terrestrial" -> TerrestrialWorldParameters.generate(config.planetType!!, config.planetSize!!, random)
			"asteroids" -> AsteroidsWorldParameters.generate(random)
			"floatingdungeon" -> FloatingDungeonWorldParameters.generate(config.dungeonWorld!!)
			else -> throw RuntimeException()
		}

		if (location.threatLevel != null) {
			visitable.threatLevel = location.threatLevel
		}

		if (config.beamUpRule != null) {
			visitable.beamUpRule = config.beamUpRule
		}

		visitable.disableDeathDrops = config.disableDeathDrops

		val template = WorldTemplate(visitable, config.skyParameters, random)
		val world = ServerWorld.create(this, template, WorldStorage.NULL, location)

		try {
			world.setProperty("ephemeral", JsonPrimitive(!config.persistent))

			if (config.useUniverseClock)
				world.sky.referenceClock = universeClock

			world.eventLoop.start()
			world.prepare(true).await()
		} catch (err: Throwable) {
			LOGGER.fatal("Exception while creating instance world at $location!", err)
			world.eventLoop.shutdown()
			throw err
		}

		return world
	}

	private suspend fun loadWorld0(location: WorldID): ServerWorld {
		return when (location) {
			is WorldID.ShipWorld -> throw IllegalArgumentException("Can't create ship worlds out of thin air")
			is WorldID.Instance -> loadInstanceWorld(location)
			is WorldID.Celestial -> loadCelestialWorld(location)
			is WorldID.Limbo -> throw IllegalArgumentException("Limbo was supplied as world ID")
		}
	}

	fun loadWorld(location: WorldID): CompletableFuture<ServerWorld> {
		return supplyAsync {
			var world = worlds[location]

			if (world != null && world.isCompletedExceptionally) {
				worlds.remove(location)
				world = null
			}

			if (world != null) {
				world
			} else {
				val future = globalScope.async { loadWorld0(location) }.asCompletableFuture()
				worlds[location] = future
				future
			}
		}.thenCompose { it }
	}

	fun loadShipWorld(connection: ServerConnection, storage: WorldStorage): CompletableFuture<ServerWorld> {
		return supplyAsync {
			val id = WorldID.ShipWorld(connection.uuid ?: throw NullPointerException("Connection UUID is null"))
			val existing = worlds[id]

			if (existing != null)
				throw IllegalStateException("Already has $id!")

			val world = ServerWorld.load(this, storage, id)
			worlds[id] = world
			world
		}.thenCompose { it }
	}

	fun notifyWorldUnloaded(worldID: WorldID) {
		execute {
			worlds.remove(worldID)
		}
	}

	init {
		scheduleAtFixedRate(Runnable {
			channels.broadcast(UniverseTimeUpdatePacket(universeClock.time), false)
			setMetadata("universe_clock", JsonPrimitive(universeClock.time))
		}, Globals.universeServer.clockUpdatePacketInterval, Globals.universeServer.clockUpdatePacketInterval, TimeUnit.MILLISECONDS)

		scheduleWithFixedDelay(Runnable {
			database.commit()
		}, Globals.universeServer.universeStorageInterval, Globals.universeServer.universeStorageInterval, TimeUnit.MILLISECONDS)

		scheduleAtFixedRate(Runnable {
			tickNormal(Starbound.TIMESTEP)
		}, Starbound.TIMESTEP_NANOS, Starbound.TIMESTEP_NANOS, TimeUnit.NANOSECONDS)

		scheduleAtFixedRate(Runnable {
			tickSystemWorlds(Starbound.SYSTEM_WORLD_TIMESTEP)
		}, Starbound.SYSTEM_WORLD_TIMESTEP_NANOS, Starbound.SYSTEM_WORLD_TIMESTEP_NANOS, TimeUnit.NANOSECONDS)

		isDaemon = false
		start()
	}

	private val occupiedNicknames = ObjectArraySet<String>()

	fun reserveNickname(name: String, alternative: String): String {
		synchronized(occupiedNicknames) {
			var name = name

			if (name.lowercase() == "server" || name.isBlank()) {
				name = alternative
			}

			while (name in occupiedNicknames) {
				name += "_"
			}

			occupiedNicknames.add(name)
			return name
		}
	}

	fun freeNickname(name: String): Boolean {
		return synchronized(occupiedNicknames) {
			occupiedNicknames.remove(name)
		}
	}

	fun clientByUUID(uuid: UUID): ServerConnection? {
		return channels.connections.firstOrNull { it.uuid == uuid }
	}

	protected abstract fun close0()
	protected abstract fun tick0(delta: Double)

	private fun tickSystemWorlds(delta: Double) {
		systemWorlds.values.removeIf {
			if (it.isCompletedExceptionally) {
				return@removeIf true
			}

			if (!it.isDone) {
				return@removeIf false
			}

			if (it.get() == null) {
				return@removeIf true
			}

			scope.launch {
				try {
					it.get()!!.tick(delta)
				} catch (err: Throwable) {
					LOGGER.fatal("Exception in system world $it event loop", err)
				}
			}

			if (it.get()!!.shouldClose()) {
				LOGGER.info("Stopping idling ${it.get()}")
				return@removeIf true
			}

			return@removeIf false
		}
	}

	private fun tickNormal(delta: Double) {
		try {
			// universeClock.nanos += Starbound.TIMESTEP_NANOS

			channels.connections.forEach {
				try {
					it.tick()
				} catch (err: Throwable) {
					LOGGER.error("Exception while ticking client connection", err)
					it.disconnect("Exception while ticking client connection: $err")
				}
			}

			tick0(delta)
		} catch (err: Throwable) {
			LOGGER.fatal("Exception in main server event loop", err)
			shutdown()
		}
	}

	override fun performShutdown() {
		super.performShutdown()

		scope.cancel("Server shutting down")
		channels.close()

		val worldSlice = ObjectArrayList(worlds.values)

		worldSlice.forEach {
			it.thenAccept { it.eventLoop.shutdown() }
		}

		worldSlice.forEach {
			it.thenAccept {
				it.eventLoop.awaitTermination(60L, TimeUnit.SECONDS)

				if (!it.eventLoop.isTerminated) {
					LOGGER.warn("World $it did not shutdown in 60 seconds, forcing termination. This might leave world in inconsistent state!")
					it.eventLoop.shutdownNow()
				}
			}
		}

		database.commit()
		database.close()
		universe.close()
		close0()
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
